@@ -200,44 +200,48 @@ gnm_compare_strings (const char *cstr1, const char *cstr2)
 {
 	const char *a = cstr1, *b = cstr2;
 	char *str1, *str2;
-	gboolean val;
+	gboolean eq;
 
-	/* If the string is identical in code points even without normalization */
-	/* then it is exactly the same. */
+	/* Skip leading ASCII prefixes that match.  */
 	while (*a == *b && *a != 0 && *b != 0)
 		a++, b++;
+	/*
+	 * If we've hit the end of one string, we ought to have hit the
+	 * end of the other.  Otherwise the strings are different.
+	 */
+	if (*a == 0 || *b == 0)
+		return *a == *b;
 
-	if (*a == 0)
-		return (*b == 0);
-	if (*b == 0)
+	/*
+	 * If they differ in two ASCII characters (including terminating
+	 * NULs), the strings must be distinct.
+	 */
+	if ((guchar)*a < 128 && (guchar)*b < 128)
 		return FALSE;
 
-	/* If they differ in two ASCII characters, the string must be distinct. */
-	if (*a > 0 && *a < 128 && *b > 0 && *b < 128)
-		return FALSE;
-
-	/* We are using NFD normalization, ie. Characters are decomposed by */
-	/* canonical equivalence, and multiple combining characters are arranged */
-	/* in a specific order. Note that ligatures remain ligatures, formatting */
-	/* such as subscript-3 versus 3 are retained. */
-
-	/* Note that for example, the distinct Unicode strings "U+212B" */
-	/*(the angstrom sign "Å") and "U+00C5" (the Swedish letter "Å") */
-	/* are both expanded by NFD (or NFKD) into the sequence */
-	/*"U+0041 U+030A" (Latin letter "A" and combining ring above "°") */
-	/* Of course "U+0041 U+030A" is retained in form, so we need to work with */
-	/* at least the last ASCII character. Performance should nearly be */
-	/* identical to using all */
-
+	/*
+	 * We are using NFD normalization, ie. Characters are decomposed by
+	 * canonical equivalence, and multiple combining characters are
+	 * arranged in a specific order. Note that ligatures remain ligatures,
+	 * formatting such as subscript-3 versus 3 are retained.
+	 *
+	 * Note that for example, the distinct Unicode strings "U+212B"
+	 * (the angstrom sign "Å") and "U+00C5" (the Swedish letter "Å")
+	 * are both expanded by NFD (or NFKD) into the sequence
+	 * "U+0041 U+030A" (Latin letter "A" and combining ring above "°")
+	 * Of course "U+0041 U+030A" is retained in form, so we need to work
+	 * with at least the last ASCII character. Performance should nearly
+	 * be identical to using all
+	 */
 	str1 = g_utf8_normalize (cstr1, -1, G_NORMALIZE_DEFAULT);
 	str2 = g_utf8_normalize (cstr2, -1, G_NORMALIZE_DEFAULT);
 
-	val = (g_strcmp0 (str1, str2) == 0);
+	eq = (g_strcmp0 (str1, str2) == 0);
 
 	g_free (str1);
 	g_free (str2);
 
-	return val;
+	return eq;
 }
 
 static GnmFuncHelp const help_exact[] = {
@@ -597,13 +601,13 @@ gnumeric_upper (GnmFuncEvalInfo *ei, GnmValue const * const *argv)
 /***************************************************************************/
 
 static GnmFuncHelp const help_concatenate[] = {
-        { GNM_FUNC_HELP_NAME, F_("CONCATENATE:the concatenation of the strings @{s1}, @{s2},\xe2\x80\xa6")},
-        { GNM_FUNC_HELP_ARG, F_("s1:first string")},
-        { GNM_FUNC_HELP_ARG, F_("s2:second string")},
+	{ GNM_FUNC_HELP_NAME, F_("CONCATENATE:the concatenation of the strings @{s1}, @{s2},\xe2\x80\xa6")},
+	{ GNM_FUNC_HELP_ARG, F_("s1:first string")},
+	{ GNM_FUNC_HELP_ARG, F_("s2:second string")},
 	{ GNM_FUNC_HELP_EXCEL, F_("This function is Excel compatible.") },
-        { GNM_FUNC_HELP_EXAMPLES, "=CONCATENATE(\"aa\",\"bb\")" },
-        { GNM_FUNC_HELP_SEEALSO, "LEFT,MID,RIGHT"},
-        { GNM_FUNC_HELP_END}
+	{ GNM_FUNC_HELP_EXAMPLES, "=CONCATENATE(\"aa\",\"bb\")" },
+	{ GNM_FUNC_HELP_SEEALSO, "LEFT,MID,RIGHT"},
+	{ GNM_FUNC_HELP_END}
 };
 
 
@@ -612,10 +616,109 @@ gnumeric_concatenate (GnmFuncEvalInfo *ei, int argc, GnmExprConstPtr const *argv
 {
 	return string_range_function (argc, argv, ei,
 				      range_concatenate,
+				      NULL,
 				      COLLECT_IGNORE_BLANKS,
 				      GNM_ERROR_VALUE);
 }
 
+static GnmFuncHelp const help_concat[] = {
+	{ GNM_FUNC_HELP_NAME, F_("CONCAT:the concatenation of the strings @{s1}, @{s2},\xe2\x80\xa6")},
+	{ GNM_FUNC_HELP_ARG, F_("s1:first string")},
+	{ GNM_FUNC_HELP_ARG, F_("s2:second string")},
+	{ GNM_FUNC_HELP_EXCEL, F_("This function is Excel compatible.") },
+	{ GNM_FUNC_HELP_EXAMPLES, "=CONCAT(\"aa\",\"bb\")" },
+	{ GNM_FUNC_HELP_NOTE, F_("This function is identical to CONCATENATE") },
+	{ GNM_FUNC_HELP_SEEALSO, "LEFT,MID,RIGHT"},
+	{ GNM_FUNC_HELP_END}
+};
+
+static GnmValue *
+gnumeric_concat (GnmFuncEvalInfo *ei, int argc, GnmExprConstPtr const *argv)
+{
+	return gnumeric_concatenate (ei, argc, argv);
+}
+
+/***************************************************************************/
+
+static GnmFuncHelp const help_textjoin[] = {
+	{ GNM_FUNC_HELP_NAME, F_("TEXTJOIN:the concatenation of the strings @{s1}, @{s2},\xe2\x80\xa6 delimited by @{del}")},
+	{ GNM_FUNC_HELP_ARG, F_("del:delimiter")},
+	{ GNM_FUNC_HELP_ARG, F_("blank:ignore blanks")},
+	{ GNM_FUNC_HELP_ARG, F_("s1:first string")},
+	{ GNM_FUNC_HELP_ARG, F_("s2:second string")},
+	{ GNM_FUNC_HELP_EXCEL, F_("This function is Excel compatible.") },
+	{ GNM_FUNC_HELP_EXAMPLES, "=TEXTJOIN(\"::\",FALSE,\"aa\",\"bb\")" },
+	{ GNM_FUNC_HELP_SEEALSO, "CONCATENATE"},
+	{ GNM_FUNC_HELP_END}
+};
+
+struct cb_textjoin {
+	char *delim;
+	gboolean ignore_blanks;
+};
+
+static int
+range_textjoin (GPtrArray *data, char **pres, gpointer user_)
+{
+	struct cb_textjoin *user = user_;
+	GString *res = g_string_new (NULL);
+	gboolean first = TRUE;
+	unsigned ui;
+
+	for (ui = 0; ui < data->len; ui++) {
+		const char *s = g_ptr_array_index (data, ui);
+
+		if (s[0] == 0 && user->ignore_blanks)
+			continue;
+
+		if (first)
+			first = FALSE;
+		else
+			g_string_append (res, user->delim);
+
+		g_string_append (res, s);
+	}
+
+	*pres = g_string_free (res, FALSE);
+	return 0;
+}
+
+static GnmValue *
+gnumeric_textjoin (GnmFuncEvalInfo *ei, int argc, GnmExprConstPtr const *argv)
+{
+	GnmValue *v;
+	gboolean err;
+	struct cb_textjoin data;
+
+	data.delim = NULL;
+
+	if (argc < 3)
+		return value_new_error_VALUE (ei->pos);
+
+	v = gnm_expr_eval (argv[0], ei->pos, GNM_EXPR_EVAL_SCALAR_NON_EMPTY);
+	if (VALUE_IS_ERROR (v))
+		goto done;
+	data.delim = value_get_as_string (v);
+	value_release (v);
+
+	v = gnm_expr_eval (argv[1], ei->pos, GNM_EXPR_EVAL_SCALAR_NON_EMPTY);
+	if (VALUE_IS_ERROR (v))
+		goto done;
+	data.ignore_blanks = value_get_as_bool (v, &err); // What about err?
+	value_release (v);
+
+	v = string_range_function (argc - 2, argv + 2, ei,
+				   range_textjoin,
+				   &data,
+				   data.ignore_blanks ? COLLECT_IGNORE_BLANKS : 0,
+				   GNM_ERROR_VALUE);
+
+done:
+	g_free (data.delim);
+
+	return v;
+}
+
 /***************************************************************************/
 
 static GnmFuncHelp const help_rept[] = {
@@ -1669,6 +1772,9 @@ GnmFuncDescriptor const string_functions[] = {
         { "unicode",    "S",                         help_unicode,
 	  gnumeric_unicode, NULL, NULL, NULL,
 	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_UNIQUE_TO_GNUMERIC, GNM_FUNC_TEST_STATUS_BASIC },
+        { "concat", NULL,               help_concat,
+	  NULL, gnumeric_concat, NULL, NULL,
+	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
         { "concatenate", NULL,               help_concatenate,
 	  NULL, gnumeric_concatenate, NULL, NULL,
 	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
@@ -1747,6 +1853,9 @@ GnmFuncDescriptor const string_functions[] = {
         { "text",       "Ss",           help_text,
 	  gnumeric_text, NULL, NULL, NULL,
 	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
+        { "textjoin", NULL, help_textjoin,
+	  NULL, gnumeric_textjoin, NULL, NULL,
+	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_NO_TESTSUITE },
         { "trim",       "S",                         help_trim,
 	  gnumeric_trim, NULL, NULL, NULL,
 	  GNM_FUNC_SIMPLE, GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
